@@ -1502,6 +1502,8 @@ ensure_surface_for_window(WindowPtr window)
             xwl_screen->tearing_control_manager, xwl_window->surface);
     }
 
+    xwl_window_set_input_region(xwl_window, wInputShape(window));
+
     return xwl_window;
 
 err:
@@ -1971,6 +1973,34 @@ xwl_window_post_damage(struct xwl_window *xwl_window)
     DamageEmpty(window_get_damage(xwl_window->surface_window));
 }
 
+void
+xwl_window_set_input_region(struct xwl_window *xwl_window,
+                            RegionPtr input_shape)
+{
+    struct wl_region *region;
+    BoxPtr box;
+    int i;
+
+    if (!input_shape) {
+        wl_surface_set_input_region(xwl_window->surface, NULL);
+        return;
+    }
+
+    region = wl_compositor_create_region(xwl_window->xwl_screen->compositor);
+    box = RegionRects(input_shape);
+
+    for (i = 0; i < RegionNumRects(input_shape); ++i, ++box) {
+        wl_region_add(region,
+                      box->x1,
+                      box->y1,
+                      box->x2 - box->x1,
+                      box->y2 - box->y1);
+    }
+
+    wl_surface_set_input_region(xwl_window->surface, region);
+    wl_region_destroy(region);
+}
+
 Bool
 xwl_window_init(void)
 {
